package PureMessage::Policy::Bundle;
use base 'PureMessage::Policy::Plugin';
use strict;
use Carp qw(croak);

use ActiveState::I18N::UTF8 qw(convert_from_utf8);
use ActiveState::UTF8 'decode_utf8';
use PureMessage::UTF8::Config qw(%utf8_config);
use PureMessage::Config qw(%config);
use PureMessage::Template qw(extract_templates);
use PureMessage::Version qw($PREFIX);
use PureMessage::Policy::Words qw(message_words);
use PureMessage::Policy::Plugin::Utils qw(get_matcher
                                          get_attachments
                                          get_recipients 
                                          set_attachments 
                                          expand_part_tmpl 
                                          recode_header
                                          set_attachment_filter_types
                                          set_attachment_filter_filetypes
                                          set_attachment_filter_extensions
                                          );
use PureMessage::_sys;
use PureMessage::Mbox::Writer;
use PureMessage::MessageStore;
use HTML::Parser;
use Fcntl ':flock';

use ActiveState::UTF8 qw(encode_rfc2047 is_valid_utf8);
use PureMessage::Sophos::SAVI::TFT qw(get_tft_mime_types
                                      get_tft_extensions
                                      get_tft_filetypes
                                      );

use PureMessage::Policy::MessageText qw(match_attachment_text
                                        per_attachment_byte_limit_exceeded
                                        per_message_time_limit_exceeded
        );

use PureMessage::Policy::Phrase::Config;

use PureMessage::MultifileCache;

use PureMessage::Policy::Bundle::Utils;

use PureMessage::Sophos::SAVI::Utils qw(set_savi_cantscan_errors);

our $VERSION = "5.601";

my $cache = PureMessage::MultifileCache->new();

my $block_size = $config{block_size};

my %commands = (
    pmx_attachment_log              => \&pmx_attachment_log,
    pmx_attachment_name             => \&pmx_attachment_name,
    pmx_attachment_size             => \&pmx_attachment_size,
    pmx_attachment_type             => \&pmx_attachment_type,
    pmx_number_of_attachments       => \&pmx_number_of_attachments,
    pmx_replace_body                => \&pmx_replace_body,
    pmx_drop_attachment             => \&pmx_drop_attachment,
    pmx_rename_attachment           => \&pmx_rename_attachment,
    pmx_add_banner                  => \&pmx_add_banner,
    pmx_convert                     => \&pmx_convert,
    pmx_set_var                     => \&pmx_set_var,
    pmx_8bit_percentage             => \&pmx_8bit_percentage,
    pmx_phrase                      => \&pmx_phrase,
    pmx_archive                     => \&pmx_archive,
    pmx_attachment_true_filetype    => \&pmx_attachment_true_filetype,
    pmx_cantscan                    => \&pmx_cantscan,
    pmx_custom_mark                 => \&pmx_custom_mark,
);

sub new {
    my $class = shift;
    my $dict = shift;
    my %opts = @_;
    return bless \%opts, $class;
}

# If PureMessage::Policy::BATV is available use it
{
    eval { require PureMessage::Policy::BATV };
}

sub commands {
    my $self = shift;

    # If the BATV package is installed allow all its commands to
    # be accessible via this package.
    eval {
        my %lookup = PureMessage::Policy::BATV->commands();
        foreach my $cmd (keys %lookup) {
            $commands{$cmd} = $lookup{$cmd};
        }
    };

    eval { 
        require PureMessage::Policy::Router;
        my %lookup = PureMessage::Policy::Router->commands();
        foreach my $cmd (keys %lookup) {
            $commands{$cmd} = $lookup{$cmd};
        }
    };

    eval { 
        require PureMessage::Policy::CreditCard;
        my %lookup = PureMessage::Policy::CreditCard->commands();
        foreach my $cmd (keys %lookup) {
            $commands{$cmd} = $lookup{$cmd};
        }
    };

    return %commands;
}

sub _str_compare {
    my $msg = shift;
    my $args = shift;

    if (my $matcher = get_matcher($msg, $args)) {
        # Note that we pass in @_ directly, so that we don't potentially make
        # a copy of big strings
        return $matcher->(@_);
    }
    else {
        croak "no match function found!";
    }
}

sub _num_compare {
    my $num = shift;
    my $args = shift;
    my $i = shift;
    $args->tag('over') and return ($num > $args->param($i));
    $args->tag('under') and return ($num < $args->param($i));
    return ($num == $args->param($i));
}

sub pmx_phrase_limit_exceeded {
    my ($msg,$args) = @_;
    return per_attachment_byte_limit_exceeded($msg)
        || per_message_time_limit_exceeded($msg);
}

sub pmx_cantscan {
    my ($msg,$args) = @_;

    $msg->log("DEBUG", "pmx_cantscan");

    return PureMessage::Policy::Bundle::Utils::cantscan($msg,$args);
}

sub pmx_phrase {
    my ($msg,$args) = @_;

    # Initialize the cantscan attribute, this will be updated if we hit any 'cantscan' limits
    $msg->set_attr('cantscan_map', [{'type' => 'phrase', 'ids' => [] }], 1);

    my $result;
    eval {
        message_words(sub {
            my (undef, undef, $words) = @_;
            my $str = join(' ', @$words);
            my @results = _str_compare($msg, $args, $str);
            if ($results[0]) {
                $result = $results[1];
                die '--matched--'; 
            }
        }, $msg, overlap => 5, per_call => 1000);
    };
    if ($@) {
        if ($@ =~ /--matched--/) {
            $msg->log('INFO', 'pmx_phrase: match found in body / text attachments');
            $msg->template("MATCHED_TEXT", $result);
            $msg->template("MATCHED_FILE", 'message_body');
            return 1;
        }
        die $@;
    }

    my $matcher = get_matcher($msg, $args);
    if ($matcher && $args->tag('attachments')) {
        my @matching_parts;
        my @non_matching_parts;
        my @matches = match_attachment_text($msg, $matcher,
            matching_parts     => \@matching_parts,
            non_matching_parts => \@non_matching_parts,
            no_early_out => $args->tag('scanall'), 
        );
        $msg->log('INFO', "pmx_phrase: match found in attachment(s) (matches=" . scalar(@matches) .")") if(@matches);

        my $minfo = $msg->minfo;
        for my $m (@matches) {
            $msg->log('INFO', "ATT: f=$m->{file} t=$m->{type} m=$m->{text}");

            $msg->template("MATCHED_TEXT", $m->{text});
            $msg->template("MATCHED_FILE", $m->{file});
            $minfo->append_items(
                    "Scanned-Attachment",
                    "$m->{type}|$m->{file}|$m->{text}",
                    );            
        }
                        
        set_attachments($msg, \@matching_parts, \@non_matching_parts);
                        
        return 1 if (@matches
                  || pmx_phrase_limit_exceeded($msg)
                  || $msg->{per_message_savi_cantscan_errors}); 
    }

    return 0;
}

sub _textify {
    my $msg = shift;
    my @txt;
    my $p = HTML::Parser->new(
        api_version => 3,
        handlers    => { text => [\@txt, 'text'] },
    );
    $msg->body(my $buf);
    $p->parse($buf);
    $p->eof;
    my $txt = '';
    for (@txt) {
        $txt .= $_ for @$_;
    }
    $txt;
}

sub pmx_attachment_log {
    my ($msg,$args) = @_;

    my (@rv, %stats, $k, $v) = ();
    my %a;


    my $cb = sub {
        my $part = shift;
        if(%a) {
            return unless $a{$part->part_id};
        }
        my $ct = $part->content_type;
        $ct =~ s/[; ].*$//; # ditch sub-typing
        $stats{$ct}||= [0,0];
#        $stats{$ct}||=$args->tag('tft') ? [0,0,''] : [0,0];
        $stats{$ct}->[0] ++;
        $stats{$ct}->[1] += $part->body_raw_size;

#        if($args->tag('tft')) {
#            my $tft_filetypes = get_tft_filetypes($part);
#            $tft_filetypes = join(':', @$tft_filetypes); 
#            $stats{$ct}->[2] .= "$tft_filetypes;";
#        }
    };
    
    # check to see if an attachment_type sieve was run, if so,
    # only process the attachments it found
    $a{$_}++ foreach (get_attachments($msg));
    $msg->enumerate_parts($cb);
        
    while(($k, $v) = each(%stats)) {
        push(@rv, join(",", @$v, $k));
    }
    
    if(@rv) {
        $msg->set_attr('mark_at', \@rv);
    }
    
    return scalar(@rv);
}

sub pmx_attachment_type {
    my ($msg,$args) = @_;
    my @matching;
    my @non_matching;

    # Initialize the cantscan attribute, this will be updated if we hit any 'cantscan' limits
    $msg->set_attr('cantscan_map', [{'type' => 'tft', 'ids' => [] }], 1);

    my @filtered = get_attachments($msg);
    $msg->enumerate_parts(sub {
        my $part = shift;
        return if @filtered && !(grep $part->part_id eq $_, @filtered);

        my @types = (scalar $part->content_type);

        my $log_string = "pmx_attachment_type (part " . $part->part_id . "):";

        if ($args->tag('tft')) { 

                # Scan and store SAVI cantscan errors
                my $tft_types = get_tft_mime_types($part);
                $msg->{per_message_savi_cantscan_errors} += set_savi_cantscan_errors($msg, 
                                                                                     $part->part_id, 
                                                                                     $part->{tft_info}->{sav_ids}->{$part->body_filename});


            if ($tft_types and @{$tft_types}) {
                push @types, @{$tft_types};
                $msg->log("DEBUG", "$log_string tft mime types: " . 
                          join(',', @$tft_types));
            } else {
                $msg->log("DEBUG", "$log_string tft mime types: none found");
            }
        }

        if (_str_compare($msg, $args, @types)) {
            push @matching, $part->part_id;
            $msg->log("DEBUG", "$log_string matched");
        }
        else {
            push @non_matching, $part->part_id;
            $msg->log("DEBUG", "$log_string didn't match");
        }
    });
    
    set_attachment_filter_types($msg, $args);

    return set_attachments($msg, \@matching, \@non_matching)
                        || $msg->{per_message_savi_cantscan_errors};

}

sub _get_names {
    my $part = shift;
    my $names = [];

    my (undef, %ct) = $part->content_type;
    push @{$names}, $ct{name} if exists $ct{name};

    my $cd = $part->header_decoded('Content-Disposition');
    push @{$names}, $1 if $cd and $cd =~ /filename\s*=\s*\"?([^\r\n;\"]*)/i;

    return $names;
}

sub pmx_attachment_name {
    my ($msg,$args) = @_;
    my @matching;
    my @non_matching;

    # Initialize the cantscan attribute, this will be updated if we hit any 'cantscan' limits
    $msg->set_attr('cantscan_map', [{'type' => 'tft', 'ids' => [] }], 1);

    my @filtered = get_attachments($msg);
    $msg->enumerate_parts(sub {
        my $part = shift;
        return if @filtered && !(grep $part->part_id eq $_, @filtered);

        my @names = @{_get_names($part)};
        my $log_string = "pmx_attachment_name (part " . $part->part_id . "):";

        if ($args->tag('tft')) {

            # Scan and store SAVI cantscan errors
            my $tft_extensions = get_tft_extensions($part);  
            $msg->{per_message_savi_cantscan_errors} += set_savi_cantscan_errors($msg, 
                                                                                 $part->part_id, 
                                                                                 $part->{tft_info}->{sav_ids}->{$part->body_filename});

            if (defined($tft_extensions) and @{$tft_extensions}) {
                push @names, @{$tft_extensions};
                $msg->log("DEBUG", "$log_string extensions: " . join(',', @$tft_extensions));
            } else {
                $msg->log("DEBUG", "$log_string extensions: none found");

            }
        } 

        if (_str_compare($msg, $args, @names)) {
            push @matching, $part->part_id;
            $msg->log("DEBUG", "$log_string matched");
        }
        else {
            push @non_matching, $part->part_id;
            $msg->log("DEBUG", "$log_string didn't match");
        }
    });
    
    set_attachment_filter_extensions($msg, $args);

    return set_attachments($msg, \@matching, \@non_matching)
                        || $msg->{per_message_savi_cantscan_errors};
}

sub pmx_attachment_true_filetype {
    my ($msg,$args) = @_;
    my @matching;
    my @non_matching;
    my @filtered = get_attachments($msg);

    # Initialize the cantscan attribute, this will be updated if we hit any 'cantscan' limits
    $msg->set_attr('cantscan_map', [{'type' => 'tft', 'ids' => [] }], 1);

    $msg->enumerate_parts(sub {
        my $part = shift;
        # If this is not an attachment, we don't care about it:
        return if @filtered && !(grep $part->part_id eq $_, @filtered);

        my $log_string = "pmx_attachment_true_filetype (part " . $part->part_id . ")";

        # Scan and store SAVI cantscan errors
        my $types = get_tft_filetypes($part);
        $msg->{per_message_savi_cantscan_errors} += set_savi_cantscan_errors($msg, 
                                                                             $part->part_id, 
                                                                             $part->{tft_info}->{sav_ids}->{$part->body_filename});

        if (defined($types) and @$types) {
            $msg->log("DEBUG", "$log_string filetypes: '" . join("', '", @$types) ."'");
        } else {
            $msg->log("DEBUG", "$log_string filetypes: none found");
        }
        
        if (_str_compare($msg, $args, @{$types})) {
            $msg->log("DEBUG", "$log_string: matched");
            push @matching, $part->part_id;
        }
        else {
            $msg->log("DEBUG", "$log_string: didn't match");
            push @non_matching, $part->part_id;
        }
    });
    
    set_attachment_filter_filetypes($msg, $args);

    return set_attachments($msg, \@matching, \@non_matching)
                        || $msg->{per_message_savi_cantscan_errors};

}

sub pmx_attachment_size {
    my ($msg,$args) = @_;
    my @matching;
    my @non_matching;

    my @filtered = get_attachments($msg);
    $msg->enumerate_parts(sub {
        my $part = shift;
        return if @filtered && !(grep $part->part_id eq $_, @filtered);

        return if $part->part_id eq '1';
        if (_num_compare($part->body_raw_size, $args, 0)) {
            push @matching, $part->part_id;
        }
        else {
            push @non_matching, $part->part_id;
        }
    });
    
    return set_attachments($msg, \@matching, \@non_matching);
}

sub pmx_number_of_attachments {
    my ($msg,$args) = @_;
    my $count = 0;
    $msg->enumerate_parts(sub {
        my $part = shift;
        return if $part->part_id eq '1';
        my $cd = $part->_content_type(lc($part->header_decoded('Content-Disposition') || 'inline'));
        return unless $cd eq 'attachment';
        $count++;
    });
    return _num_compare($count, $args, 0);
}

sub _replace_body {
    my $part = shift;
    my $args = shift;
    my ($ct, $te);
    if ($args->tag('data') && $_[0] =~ /[\x80-\xff]/) {
        # data is verbatim and contains non-ascii chars
        # all data from policy.siv is in UTF-8
        $ct = 'text/plain; charset=UTF-8';
        $te = '8bit';
    }
    else {
        $ct = 'text/plain';
        $te = '7bit';
    }
    my $tag = $args->tag('content_type');
    $ct = $tag->{params}->[0]->{value}->[0] if $tag && ! ($args->tag('data'));
    $tag = $args->tag('transfer_encoding');
    $te = $tag->{params}->[0]->{value}->[0] if $tag && ! ($args->tag('data'));

    # Copy the body *before* modifying anything, otherwise expand_part_tmpl
    # may do the wrong thing.
    $part->body_copy_from(expand_part_tmpl($part, $_[0]));
    $part->header_replace('Content-Type',$ct);
    $part->header_replace('Content-Transfer-Encoding',$te);
    $part->header_replace('Content-Disposition','inline');
}

sub _get_groups {
    my ($msg, $args) = @_;
    my @groups = ();

    my $direction = $msg->minfo_value('Direction');
    my @_groups = ($direction eq 'Outbound')
        ? $msg->minfo_value('Sender-Group')
        : $msg->minfo_value('Recipient-Group');

    if(@_groups) {
        my %_groups = ();
        for my $_group (@_groups) {
            my ($recipient, $group) = split /\|/, $_group;
            # After message splitting, re-map mapped-recipients (part of the
            # entire list of recipient) to their respective groups
            $_groups{$recipient} = $group;
        }
        if($direction eq 'Outbound') {
            push @groups, (values %_groups)[0];
        }
        else {
            for my $recipient (get_recipients($msg)) {
                push @groups, $_groups{$recipient} if $_groups{$recipient};
            }
        }
    }

    return \@groups;
}


sub _get_data {
    my ($msg,$args) = @_;
    my $data = "";

    # Fetch group-specific banner first
    my $group_tag = $args->tag('groupfile');
    if($cache && $group_tag) {
        my $groups = _get_groups($msg, $args);
        if($groups && @$groups) {
            eval {
                my $multifile_id = $group_tag->{params}->[0]->{value}->[0];
                my $multifile = $cache->get_multifile($multifile_id);
                # Only use the first group because, once we hit this,
                # it means the message has been split and we only have
                # one group.
                $data = $multifile->get_encoded_content($groups->[0]);
            };
            return $data unless $@;
        }
    }

    if ($args->tag('data')){
        $data = $args->param(0)->[0];
    }
    else {
        my $file = $args->param(0)->[0];
        # 0+ as filename does not make sense
        return undef unless $file;
        
        #relative paths are interpreted relative to PREFIX
        $file = "$PREFIX/$file" unless index($file, '/') == 0;
        if (open F, $file) {
            my $str = join '', <F>;
            close F;
            $data = $msg->template_expand($str);
        }
        else {
            $msg->log("ERR", "POLICY ERROR: could not open '$file': $!");
        }
    }
    return $data;
}

sub pmx_replace_body {
    my ($msg,$args) = @_;
    
    my @attachments = get_attachments($msg);
        
    my $data = _get_data($msg, $args);
    
    if (@attachments) {
        my $count = 0;
        $msg->enumerate_parts(sub {
            my $part = shift;
            my $id = $part->part_id;
            $count++, _replace_body($part, $args, $data) if grep $id eq $_, @attachments;
        });
        $msg->log("DEBUG", "pmx_replace_body: $count part" . ($count==1 ? '' : 's') . ' replaced' );
    }
    else {
        $msg->part_delete($_) foreach ($msg->parts);
        _replace_body($msg, $args, $data);
        $msg->log("DEBUG", "pmx_replace_body: replaced main message body");
    }
}

sub pmx_drop_attachment {
    my ($msg,$args) = @_;
    
    my @attachments = get_attachments($msg);
    
    if (@attachments) {
        my $count = 0;
        $msg->enumerate_parts(sub {
            my $part = shift;
            my $id = $part->part_id;
            foreach my $attachment (@attachments) {
                if ($attachment =~ /^$id\.(\d+)$/) {
                    $part->part_delete("1.$1");
                    $count++;
                }
            }
        });
        $msg->log("DEBUG", "pmx_drop_attachment: $count attachment" . ($count==1 ? '' : 's') . ' dropped' );
    }
    else {
        $msg->log("DEBUG", "pmx_drop_attachment: no attachments dropped");
        # XXX for now, do nothing (drop the whole body maybe?  might be dangerous)
    }
}

my $re = {
    ct => 'name\s*=\s*\"?([^\r\n;"]*)\"?',
    cd => 'filename\s*=\s*\"?([^\r\n;"]*)\"?',
};
sub pmx_rename_attachment {
    my ($msg, $args) = @_;

    my @attachments = get_attachments($msg);
    # XXX encode_rfc2047() sucks.
    # It considers all high-bit-set characters as Latin1.
    my $tmpl = encode_rfc2047($args->param_decoded(0)->[0]);
    my $tmpl_flag = ($tmpl =~ /%%ATTACHMENT_NAME%%/) ? 1 : 0;

    return unless @attachments;

	my $count = 0;
	$msg->enumerate_parts(sub {
	    my $part = shift;
	    my $id = $part->part_id;

	    if(grep $id eq $_, @attachments) {
            # XXX header() can be replaced by header_decoded(), which can
            # decode rfc2047 headers, but the counterpart to encode in
            # header_replace() can only encode into ISO-8859-1 encoded
            # header.
            # XXX Tested with Outlook Express 6 & Thunderbird 1.0, it's
            # okay to add plain words around encoded words.
            my @ct = $part->header('Content-Type');
            my @cd = $part->header('Content-Disposition');

            my ($ct_flag, $cd_flag) = (0, 0);

            my $i = 0;
            foreach (@ct) {
                if($_ and $_ =~ /$re->{ct}/io) {
                    my ($tmp, $name) = ($1, $tmpl);
                    $name =~ s/%%ATTACHMENT_NAME%%/$tmp/g if $tmpl_flag;

                    $_ =~ s/$re->{ct}/name="$name"/io;
                    $part->header_replace("Content-Type", $_, $i);

                    $ct_flag = 1 unless $ct_flag;
                }
                $i++;
            }

            $i = 0;
            foreach (@cd) {
                if($_ and $_ =~ /$re->{cd}/io) {
                    my ($tmp, $name) = ($1, $tmpl);
                    $name =~ s/%%ATTACHMENT_NAME%%/$tmp/g if $tmpl_flag;

                    $_ =~ s/$re->{cd}/filename="$name"/io;
                    $part->header_replace("Content-Disposition", $_, $i);

                    $cd_flag = 1 unless $cd_flag;
                }
                $i++;
            }

            $count++ if $ct_flag or $cd_flag;
	    }
	});
	$msg->log("DEBUG", "pmx_rename_attachment: $count attachment" . ($count == 1 ? '' : 's') . ' renamed');
}

# from my experiments with the current AddBanner
# i interpret its behaviour as:
# look for a text/html or text/plain mime part
# if you find one, append the banner.
# [NeilW] If not, add a new text/plain part and insert the banner.

my $re_html_top = ['<body[^>]*>', '<html[^>]*>', '^'];
my $re_html_bottom = ['</body', '</html', '$'];
sub pmx_add_banner {
    my ($msg,$args) = @_;

    my $banner_data = _get_data($msg, $args);
    # '0+' as banner does not make sense
    return unless $banner_data;

    my $body = $args->tag('body');
    my $charset = $args->tag('charset');
    my $header = $args->tag('header');
    my $ignore = $args->tag('ignore');
    my $top = $args->tag('top');
    my %ignore;
    %ignore = map { $_ => 1 } @{ $ignore->{params}[0]{value} } if $ignore;

    if ($body or !$header) {
        my $found = 0;
        $msg->enumerate_parts(sub {
            my $part = shift;
            my ($ct, %ct_p) = $part->content_type();
            my $data = $banner_data;

            if ($ignore{$ct}) {
                $found = 1;
                return;
            }
                
            # Skip attachments, even if we haven't been told to ignore them.
            for my $disposition ($part->header("Content-Disposition")) {
                # XXX This code may break if filename contains 'attachment'.
                # See the code in pmx_number_of_attachments() for similar things.
                return if $disposition =~ /\battachment\b/i;
            }
            
            if($charset and $ct_p{charset} ) {
                my $charset_val = $charset->{params}->[0]->{value}->[0];
                
                # Only continue if the requested charset
                # matches the charset of this message part.
                return unless $ct_p{charset} =~ /^$charset_val$/io;
                
                # If we have verbatim text, we may need 
                # to convert it to the requested charset. 
                if ($args->tag('data')) {
                    if ($charset_val 
                            =~ /(\basc[-_]?ii\b)|(\biso[-_]?8859[-_]1\b)/i) {
                        # If the charset is ascii or iso-8859-1,
                        # we need to do the latin1 conversion.
                        $data = decode_utf8($data);
                    }
                    elsif ($charset_val !~ /\butf[-_]?8/i) {
                        # For all other charsets (except utf-8), we need
                        # to do the non-latin1 conversions (using iconv).
                        my $result = convert_from_utf8(
                            $data, $ct_p{charset}, \%utf8_config);
                        unless (defined $result) {
                            $msg->log("DEBUG", 
                                "pmx_add_banner: "
                                . "failed to convert banner text to charset: " 
                                . $ct_p{charset} 
                                . " - is iconv enabled in utf8.conf?");
                            return;
                        }
                        $data = $result;
                    }                                                    
                }
            }

            if($ct eq 'text/plain') {
                if($top) {
                    my $body;
                    $part->body($body);
                    $part->body_copy_from("$data\n" . $body);
                }
                else {
                    $part->body_append("\n$data\n");
                }
                if ( $data =~ /[\x80-\xff]/) {  # data is verbatim and contains non-ascii chars
                    $part->header_replace("Content-Transfer-Encoding", "8bit");
                }
                $msg->log("DEBUG", "pmx_add_banner: add banner to '$ct' part");
                $found = 1;
                $ignore{$ct} = 1;
            }
            if($ct eq 'text/html') {
                my $body;
                $part->body($body);
                $data = "<PRE>$data</PRE>" if $args->tag('use_html_pre');

                if($top) {
                    foreach (@$re_html_top) {
                        last if $body =~ s/($_)/$1$data/i;
                    }
                }
                else {
                    foreach (@$re_html_bottom) {
                        last if $body =~ s/($_)/$data$1/i;
                    }
                }

                $part->body_copy_from($body);
                if ( $data =~ /[\x80-\xff]/) {  # data is verbatim and contains non-ascii chars
                    $part->header_replace("Content-Transfer-Encoding", "8bit");
                }

                $msg->log("DEBUG", "pmx_add_banner: add banner to '$ct' part");
                $found = 1;
                $ignore{$ct} = 1;
            }
        });
        if (!$found) {
            $msg->make_mixed();
            my $part = $msg->part_add();
            $part->header_add('Content-Type', "text/plain; charset=UTF-8");
            $part->header_add('Content-Disposition', 'inline');
            $part->body_append("\n$banner_data\n");
            if ( $banner_data =~ /[\x80-\xff]/) {  # data is verbatim and contains non-ascii chars
                $part->header_replace("Content-Transfer-Encoding", "8bit");
            }
            $msg->log("DEBUG", "pmx_add_banner: add banner to newly created 'text/plain' part");
        }
    }

    my $name = $header->{params}->[0]->{value}->[0] if $header;
    if ($name) {
        # force latin1 here as the Header option only works with latin1.
        my $data = decode_utf8($banner_data);
        recode_header($data) if $args->tag('recode');
        $data =~ s/\n/\n\t/g;
        $msg->header_add($name, $data);
        $msg->log("DEBUG", "pmx_add_banner: added banner as '$name' header");
    }
}

sub pmx_convert {
    my ($msg,$args) = @_;
    croak "not implemented";
}

sub pmx_set_var {
    my ($msg,$args) = @_;
    
    my $template = $args->param(0)->[0];
    my $command = $args->param(1)->[0];
    my @list = extract_templates($command);
    if (grep /^MESSAGE_FILE$/i, @list) {
        $msg->template('MESSAGE_FILE', $msg->filename);
    }
    my $expanded = $msg->template_expand($command);
    chomp (my $output = `$expanded`);
    $msg->template($template, $output);
}

sub pmx_8bit_percentage {
    my ($msg,$args) = @_;
    $msg->assert_event('eom');
    my $offset = 0;
    my $buf;
    my $count;
    while (my $read = $msg->message_raw($buf,$offset, 2048)) {
        $read ? $offset += $read : last;
        $count +=  $buf =~ tr/\200-\377//;
    }
    my $percent = int($count/$msg->message_size*100);
    return _num_compare($percent, $args, 0);
}

sub pmx_archive {
    my ($msg,$args) = @_;
    $msg->assert_event('eom');
    my $to_file = $args->param(0)->[0];
    
    my ($writer, $closer);
    
    if ($args->tag('mstore')) {
        # support scheme other than 'dir'
        my $scheme = 'dir';
        if ($to_file =~ s,^(\w+):,,) {
            $scheme = $1;
        }
        # absolutize path if necessary
        unless ($to_file =~ m,^/,,) {
            $to_file = "$ENV{HOME}/$to_file" if defined $ENV{HOME};
        }
        # now we should have valid messagestore id.
        $to_file = "$scheme:$to_file";
        my $mstore = PureMessage::MessageStore->new($to_file, create=>1);
        my $uniq = PureMessage::_sys::tid() . (defined($msg->tmp_id()) ? "-" . $msg->tmp_id() : "");
        my $tmpfile = $mstore->new_tempfile($uniq);
        my $tmpfilename = $tmpfile->filename;
        $writer = sub {$tmpfile->print($_[0]) or die "Error while appending to file '$tmpfilename': $!"};
        $closer = sub {$tmpfile->store};
    }
    else {
        unless ($to_file =~ m,^/,,) {
            $to_file = "$ENV{HOME}/$to_file" if defined $ENV{HOME};
        }
        open my $fh, ">>", $to_file or die "Could not open file '$to_file' for writing: $!";
        chmod 0600, $to_file;
        flock($fh, LOCK_EX) or die "Could not lock '$to_file': $!";
        my $mbox = PureMessage::Mbox::Writer->new($fh, block_size=>$block_size);
        $mbox->from(env_sender => scalar $msg->minfo_value('Envelope-From'));
        $writer = sub {$mbox->data($_[0]) or die "Error while appending to file '$to_file': $!"};
        $closer = sub {
            $mbox->eom or die "Error appending to file '$to_file': $!";
            flock($fh, LOCK_UN);
            close($fh) or die "Error closing file '$to_file': $!";
        };
    }
    
    if ($args->tag('original')) {
        my $offset;
        while (my $read = $msg->message_raw(my $buf, $offset, $block_size)) {
            $read ? $offset += $read : last;
            $writer->($buf);
        }
    }
    else {
        $msg->implode(sub{
            $writer->(join('', @_));
        });    
    }
	
    $closer->();	
}

sub pmx_custom_mark {
    my $msg = shift;
    my $args = shift;
    my $name = $args->param(0)->[0];
    my $value = $args->param(1)->[0];
    my $attr;

    my $MULTI  = "xmmark_crpt_$name";
    my $SINGLE = "x1mark_crpt_$name";
    if ($args->tag('overwrite')) {
        $attr = [$value];
        $msg->set_attr($SINGLE, $attr);
        $msg->del_attr($MULTI);
    }
    else {
        $attr = $msg->del_attr($SINGLE) || $msg->get_attr($MULTI) || [];
        push @{$attr}, $value;
        $msg->set_attr($MULTI, $attr);
    }

    $msg->log("DEBUG", "pmx_custom_mark: '$name':'$value' " .
              ($args->tag('overwrite') ? '(overwrite)' : '(append)'));
}

1;
